/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_rstrt_req.c,v 1.123 2005/08/25 04:41:00 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/errno.h>
#include <linux/pipe_fs_i.h>
#include <linux/time.h>
#include <linux/mman.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#include "cr_context.h"
#include "vmadump.h"

#ifndef PID_MAX
   /* XXX: Is this always true? */
  #define PID_MAX 0x8000
#endif


/*
 * alloc_rstrt_request
 */
static cr_rstrt_req_t *
alloc_rstrt_req(void)
{
    cr_rstrt_req_t *req = NULL;
    cr_objectmap_t map;

    map = cr_alloc_objectmap();
    if (!map) {
	goto out;
    }

    req = kmalloc(sizeof(*req), GFP_ATOMIC);
    if (req) {
        MOD_INC_USE_COUNT;
        memset(req, 0, sizeof(*req));
        atomic_set(&req->ref_count, 1);
	req->linkage = NULL;
	req->cr_restart_task = current;
	get_task_struct(req->cr_restart_task);
	set_bit(0, &req->need_children);
	init_MUTEX(&req->serial_mutex);
	init_waitqueue_head(&req->wait);
	req->result = 0;
	req->die = 0;
	req->map = map;
    } else {
	goto out_freemap;
    }

out:
    return req;

out_freemap:
    cr_release_objectmap(map);
    return req;
}

static void
release_rstrt_req(cr_rstrt_req_t *req)
{
    CR_KTRACE_REFCNT("ref count is approximately %d", 
		     atomic_read(&req->ref_count));

    if (atomic_dec_and_test(&req->ref_count)) {
        CR_KTRACE_REFCNT("fput()'ing file");
	fput(req->file);
	if (req->linkage) {
	    CR_KTRACE_ALLOC("kfree()'ing linkage");
	    kfree(req->linkage);
	}
	put_task_struct(req->cr_restart_task);
	CR_KTRACE_ALLOC("kfree()'ing req");
	cr_release_objectmap(req->map);
        kfree(req);
        MOD_DEC_USE_COUNT;
    } else {
	clear_bit(0, &req->need_children);
	wake_up(&req->wait);
    }
}

static int
make_task_req(void)
{
    return -ENOSYS;
}

static int
make_proc_req(void)
{
    return -ENOSYS;
}

/*
 * source get typed.
 *
 * Doesn't quite get concurrency.
 */
static struct file *
__cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return req->file;
}

/*
 * cr_src_gett
 *
 * source get typed.  Knows what we're planning to write, and understands
 * how to serialize when necessary.
 */
static struct file *
cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return __cr_src_gett(req, object);
}

/*
 * __cr_src_putt
 *
 * stands for src put typed.
 */
static int
__cr_src_putt(struct file *filp)
{
    return 0;
}

/*
 * cr_src_putt
 *
 * stands for src put typed.
 */
static int
cr_src_putt(struct file *filp)
{
    return __cr_src_putt(filp);
}

static int cr_restore_creds(cr_rstrt_req_t *req)
{
    struct cr_context_creds cf_creds;
    int retval;

    retval = cr_kread(req->file, &cf_creds, sizeof(cf_creds));
    if (retval < sizeof(cf_creds)) {
	retval = -EIO;
        goto out;
    }
     
#if CR_RESTORE_IDS
    {
	int i, j;
	if (cf_creds.ngroups > NGROUPS) {
	    CR_KTRACE_LOW_LVL("%s: Invalid supplemental group count (%d)",
			      __FUNCTION__, (int)cf_creds.groups);
	    retval = -EINVAL;
	    goto out;
	}
	/* Search for any "expansion" of the group set */
	for (i = 0; i < cf_creds.groups; ++i) {
	    gid_t g = cf_creds.groups[i];
	    int gid_ok = 1;
	    for (j = 0; j < current->ngroups; ++j) {
		if (g == current->groups[j]) {
		    gid_ok = 0;
		    break;
		}
	    }
	    if (!gid_ok) {
	        /* If we reach here then we've seen a supplemental group in the
	         * saved context, which is not present in the current list.
		 * The set_groups() call checks permissions for us.  If we fail,
		 * then we must not have enough credentials.
		 */
		mm_segment_t oldfs = get_fs();
		set_fs(KERNEL_DS);
		retval = sys_setgroups(cf_creds.ngroups, cf_creds.groups);
		set_fs(oldfs);
		if (retval < 0) {
		    CR_KTRACE_LOW_LVL("%s: Failed to restore supplemental groups", __FUNCTION__);
		    goto out;
		}
		CR_KTRACE_LOW_LVL("%s: Restored supplemental groups", __FUNCTION__);
		break; /* no need to continue the i loop */
	    }
	}

	/* The set_setresgid() call checks permissions for us, always OK if no change . */
	retval = sys_setresgid(cf_creds.gid, cf_creds.egid, cf_creds.sgid);
	if (retval < 0) {
	    CR_KTRACE_LOW_LVL("%s: Failed to restore gids", __FUNCTION__);
	    goto out;
	}
	CR_KTRACE_LOW_LVL("%s: Restored gids", __FUNCTION__);

	/* The set_setresuid() call checks permissions for us, always OK if no change . */
	retval = sys_setresuid(cf_creds.uid, cf_creds.euid, cf_creds.suid);
	if (retval < 0) {
	    CR_KTRACE_LOW_LVL("%s: Failed to restore uids", __FUNCTION__);
	    goto out;
	}
	CR_KTRACE_LOW_LVL("%s: Restored uids", __FUNCTION__);
    }
#endif	 /* CR_RESTORE_IDS */

out:
    return retval;
}

static int
cr_linkage_make_restart_task(struct task_struct *cr_restart_task,
                             struct cr_context_tasklinkage *l)
{
    /* We don't want any myptr's to match this by accident... */
    l->myptr = NULL;
    l->real_parent = NULL;
    l->parent = NULL;

#if HAVE_2_4_TASK_LINKAGE
    INIT_LIST_HEAD(&l->thread_group);
    /* Probably won't need that either */
#endif

    l->pid = cr_restart_task->pid;
    l->tgid = cr_restart_task->tgid;
    l->pgrp = cr_task_pgrp(cr_restart_task);
    l->tty_old_pgrp = cr_task_tty_old_pgrp(cr_restart_task);
    l->session = cr_task_session(cr_restart_task);
    l->leader = cr_task_leader(cr_restart_task);
    /* 
     * Or any of that stuff
     */

    l->tl_task = cr_restart_task;
    /*
     * But we need that one
     */

    return 0;
}

static int
cr_linkage_make_init_task(struct task_struct *task,
                          struct cr_context_tasklinkage *l)
{

    if (task == NULL) {
        CR_ERR("%s [%d]:  ERROR:  No task found!", __FUNCTION__, current->pid);
        return -1;
    }

    if (l == NULL) {
        CR_ERR("%s [%d]:  ERROR:  NULL linkage!", __FUNCTION__, current->pid);
        return -EINVAL;
    }

    l->myptr = CR_PARENT_INIT;
    l->tl_task  = task;
    l->pid   = 1;

    /* XXX: not clear to me (PHH) that anything below this point is needed AT ALL */

    /*
     * This will make sure our links work correctly...
     * XXX:  But we need to make sure that appending the lists together
     * is handled specially at the end...
     */
    l->real_parent = NULL;
    l->parent = NULL;

#if HAVE_2_4_TASK_LINKAGE
    INIT_LIST_HEAD(&l->thread_group);
#endif

    l->tgid = task->tgid;
    l->pgrp = cr_task_pgrp(task);
    l->tty_old_pgrp = cr_task_tty_old_pgrp(task);
    l->session = cr_task_session(task);
    l->leader = cr_task_leader(task);

    return 0;
}

/*
 * cr_load_linkage
 *
 * Loads the linkage structure stored in the context file into a new
 * buffer attatched to request
 *
 * This linkage structure is read, later, by cr_restore_linkage
 *
 * This grabs and releases tasklist_lock.
 *
 * NOTE:  Can use this later to reconstruct the graph of processes,
 * to restore the session and thread counts
 */
static int cr_load_linkage(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *linkage;
    int linkage_bytes;                  /* size of linkage stored on disk */
    int linkage_size;                   /* number of entries in array */
    int retval;
    int i;

    /* find out how large the linkage structure is */
    retval = cr_kread(req->file, &linkage_bytes, sizeof(linkage_bytes));
    if (retval < sizeof(linkage_bytes)) {
        goto out;
    }

    /* allocate memory for the linkage structure + 2 extras */
    linkage = NULL;
    linkage_size = 0;
    if (linkage_bytes + 2*sizeof(*linkage) < 131072) {
        linkage = kmalloc(linkage_bytes + 2*sizeof(*linkage), GFP_KERNEL);
	linkage_size = linkage_bytes/sizeof(*linkage) + 2;
    }
    if (linkage == NULL || linkage_size == 0) {
        CR_WARN("Too many processes (size = %d)", linkage_bytes);
        retval = -ENOMEM;
	goto out;
    }

    /* Now load the linkage from disk */
    /* NOTE:  The first two entries are used later for cr_restart and
     * init */
    retval = cr_kread(req->file, &linkage[2], linkage_bytes);
    if (retval < linkage_bytes) {
        kfree(linkage);
        goto out;
    }

    /* 
     * Initialize all tl_task's to NULL
     *
     * Note:  Do not mess skip this.  We use it to check for process aborts in 
     * cr_do_restore_linkage.  See cr_restore_linkage.
     */
    for (i=0; i<linkage_size; ++i) {
        linkage[i].tl_task = NULL;
    }

    /* Add linkage entry for cr_restart task */
    cr_linkage_make_restart_task(req->cr_restart_task, &linkage[0]);
    /* Add linkage entry for init */
    read_lock(&tasklist_lock);
    cr_linkage_make_init_task(find_task_by_pid(1), &linkage[1]);
    read_unlock(&tasklist_lock);

    /* attatch the linkage structure to the request */
    req->linkage = linkage;
    req->linkage_size = linkage_size;
    retval = 0;

out:
    return retval;
}

#if HAVE_2_6_TASK_LINKAGE
/* Code based heavily on kernel/pid.c */
#define BITS_PER_PAGE (8*PAGE_SIZE)
#define BITS_PER_PAGE_MASK (BITS_PER_PAGE-1)
typedef struct pidmap {
        atomic_t nr_free;
        void *page;
} pidmap_t;
#if CR_KDATA_pidmap_array
  extern pidmap_t pidmap_array[];
#endif
#if CR_KDATA_pidmap_lock
  extern spinlock_t pidmap_lock;
#endif
static inline void
_cr_free_pid(int pid)
{
    pidmap_t *map = &pidmap_array[pid / BITS_PER_PAGE];
    int offset = pid & BITS_PER_PAGE_MASK;

    clear_bit(offset, map->page);
    atomic_inc(&map->nr_free);
}
static int
_cr_alloc_pid(int pid) /* MUST call with no locks held */
{
    pidmap_t *map = &pidmap_array[pid / BITS_PER_PAGE];
    int offset = (pid & BITS_PER_PAGE_MASK);
    int result = 0;	/* assume failure */

    /* make sure we have a page */
    if (unlikely(map->page == NULL)) {
	unsigned long page = get_zeroed_page(GFP_KERNEL);

	spin_lock(&pidmap_lock);
	if (map->page) {
	    /* lost a race, free our page */
	    free_page(page);
	} else {
	    map->page = (void *)page;
	}
	spin_unlock(&pidmap_lock);
    }
    /* XXX assert (map->page != NULL) */

    if (!test_and_set_bit(offset, map->page)) {
	atomic_dec(&map->nr_free);
	result = 1;
    }

    return result;
}

/* Reserve all the pids we need
 * Returns non-zero on failure
 */
static int
cr_reserve_ids(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    CR_KTRACE_HIGH_LVL("Now reserving required ids... ");

    /* Don't need to worry about session IDs, since exactly one member
     * of the session, the leader, will have a matching PID.  (When that
     * leader exits the entire session dies).
     */

#if 0 /* Not yet */
    /* _DO_ need to think about PGIDs.  Since it is possible for the process
     * which originally had pid==pgid to exit w/o terminating the process group,
     * we must call _cr_alloc_pid() for any such ids.  Easy way is to do this
     * for every pgid we restore and filter the pid==pgid cases in the pid loop.
     */
    foreach_unique_pgid_req {
        if (!_cr_alloc_pid(pgid)) {
            goto out_release_pgids;
        }
    }
#endif
    
    for (i = 2; i < req->linkage_size; ++i) {
	link = &req->linkage[i];

#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we already grabbed this id in the pgid loop above */
	    continue;
	}
#endif

	if (!_cr_alloc_pid(link->pid)) {
#if CR_KERNEL_TRACING
	    read_lock(&tasklist_lock);
	    printk("Found pid %d in use: pid=%p, tgid=%p, pgid=%p, sid=%p\n",
		   link->pid,
		   find_pid(PIDTYPE_PID, link->pid),
		   find_pid(PIDTYPE_TGID, link->pid),
		   find_pid(PIDTYPE_PGID, link->pid),
		   find_pid(PIDTYPE_SID, link->pid));
	    read_unlock(&tasklist_lock);
#endif
	    goto out_release_pids;
	}
    }

    return 0;	/* success */

out_release_pids:
    /* If we failed w/ i=N, then 2<=j<N were allocated */
    while (i>2) {
	--i;
	_cr_free_pid(req->linkage[i].pid);
    }
#if 0	/* Not yet */
out_release_pgids:
    XXX release whichever pgids we had obtained before failure.
#endif

    return 1;	/* failure */
}

static void
cr_release_ids(cr_rstrt_req_t *req)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    for (i = 2; i < req->linkage_size; ++i) {
	link = &req->linkage[i];
#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we will free this id in the pgid loop below */
	    continue;
	}
#endif
	_cr_free_pid(link->pid);
    }

#if 0	/* Not yet */
    foreach_unique_pgid_req {
	_cr_free_pid(pgid);
    }
#endif
}
#endif /* HAVE_2_6_TASK_LINKAGE */

/*
 * cr_is_linkage_restorable
 *
 * Scans through a linkage and checks to see if any of the PIDs, etc.
 * that are described here are in use.  (i.e. check for PID conflict)
 * Also check for missing (NULL) linkage fields.
 *
 * NOTE:  Called from request_restart _AND_ from cr_do_restore_linkage!
 * As such, this thing _might_ be called with a spinlock held, so be
 * careful with what you go around doing in here, as you may get in trouble.
 *
 * NOTE: cr_is_linkage_restorable caller must read_lock tasklist_lock
 */
static int
cr_is_linkage_restorable(cr_rstrt_req_t *req)
{
    int retval;
    int i;
    struct cr_context_tasklinkage *l;
#if HAVE_2_4_TASK_LINKAGE
    struct task_struct *p;
#endif

    retval = 0;

    /* Make sure that cr_restart_task is around and initialized properly */
    if (req->linkage[0].tl_task == NULL) {
        CR_ERR("%s [%d]:  NULL cr_restart_task!", __FUNCTION__, current->pid);
	return 0;
    }

    /* Try to avoid a possible race against req->cr_restart_task exiting
     * before we get to restore the linkage */
    if (cri_task_dead(req->linkage[0].tl_task)) {
        CR_ERR("%s [%d]:  Unrestorable linkage: cr_restart_task exited "
		"prematurely!", __FUNCTION__, current->pid);
        return 0;
    }

    for(i=2; i<req->linkage_size; ++i) {
	l = &req->linkage[i];
	
	if (!l->parent || !l->real_parent) {
	    CR_WARN("parent(s) not found ... i=%d", i);
            return 0;
	}
#if HAVE_2_4_TASK_LINKAGE
        if (l->pid >= PID_MAX) {
	    CR_WARN("PID out of bounds ... i=%d pid=%d", i, l->pid);
            return 0;
	}
        for_each_task(p) {
            if (l->pid == p->pid || l->tgid == p->tgid) {
		/* XXX: sessions not supported yet...
	        l->pgrp == p->pgrp || 
		l->sesion == p->session
		*/
	        CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
                return 0;
	    }
	}
#else
	// XXX: should he validate some equivalent of PID_MAX ?
	if (!l->group_leader) {
	    CR_WARN("thread group leader not found ... i=%d", i);
            return 0;
	}
	if (find_pid(PIDTYPE_PID, l->pid) ||
	    find_pid(PIDTYPE_TGID, l->tgid)
	    /* XXX: sessions not supported yet...
	    find_pid(PIDTYPE_PGID, l->pgrp) ||
	    find_pid(PIDTYPE_SID, l->session) */
	   ) {
	    CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
            return 0;
	}
#endif
    }

    return 1;
}

// cr_rstrt_req(user_req)
//
// Processes a restart request received from user space.
//
// Returns 0 on sucess
// Returns negative on failure
// XXX: Return number of children to start when doing process groups?
int cr_rstrt_request_restart(struct file *filp, void *arg)
{
    struct cr_context_header cf_header;
    cr_pdata_t *priv;
    struct cr_rstrt_args args;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    /* Make sure that we don't have any requests already on this fd */
    priv = filp->private_data;
    if (priv->rstrt_req || priv->chkpt_req) {
        return -EAGAIN;
    }

    /* XXX:  We don't handle a cr_restart that has children
     * yet, so make sure that you don't have any before continuing */
    if (CR_CHILD(current) != NULL) {
        CR_KTRACE_UNEXPECTED("Aborting.  Process has children.");
        return -EAGAIN;
    }

    /* Allocate a request structure for bookkeeping */
    req = alloc_rstrt_req();
    if (req == NULL) {
        return -ENOMEM;
    }
    priv->rstrt_req = req;
    
    /* Copy in the user's arguments */
    if (copy_from_user(&args, arg, sizeof(args))) {
	retval = -EFAULT;
        goto out_free_req;
    }

    // XXX:  Copy in user request options
    
    /* Check the file descriptor */
    // XXX:  Needs to be replaced with cr_rstrt_src_get()...
    req->file = fget(args.cr_fd);
    if (!(req->file)) {
	retval = -EBADF;
        goto out_free_req;
    }

    /* Read in the context file header */
    retval = cr_kread(req->file, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) goto out_free_req;

    CR_KTRACE_HIGH_LVL("cr_magic = %d %d, " 
		       "cr_version = %d, "
		       "checkpoint_type = %d, "
		       "num_threads = %d",
		       cf_header.magic[0], cf_header.magic[1], cf_header.version, 
		       cf_header.checkpoint_type, cf_header.num_threads);


    // XXX: Check cr_magic
    //
    if (cf_header.num_threads <= 0) {
        retval = -EBADSLT;
	goto out_free_req;
    }
    
    // Check cr_version
    if (cf_header.version != CR_CONTEXT_VERSION) {
        retval = -EBADRQC;
	goto out_free_req;
    }
    
    /* XXX: Have this check to make sure the type matches what the user
     * asked for. */
    switch(cf_header.checkpoint_type) {
        case CR_SCOPE_TASK:
	    retval = make_task_req();
	    break;
	case CR_SCOPE_PROC:
	    retval = make_proc_req();
	    break;
	case CR_SCOPE_PGRP:
	    retval = -ENOSYS;
	    goto out_free_req;
	    break;
	case CR_SCOPE_SESS:
	    retval = -ENOSYS;
	    goto out_free_req;
	    break;
	default:
	    retval = -EBADSLT;
	    goto out_free_req;
	    break;
    }
    
    /* Set number of threads for this one. */
    req->total_threads = cf_header.num_threads;
    req->clones_needed = cf_header.num_threads - 1;

    /* Initialize parts of req relevant to restoring task linkage */
    /* First load the linkage */
    retval = cr_load_linkage(req);
    if (retval < 0) {
        CR_WARN("Error reading linkage structure! err=%d", retval);
	retval = -ENOMEM;
	goto out_free_req;
    }

    /* Now initialize the barrier count so we can block later */
    /* XXX:  This may not be correct.  We're assuming here that all the
     * threads that we create will actually call into the barrier code.
     * While this is true in all sane cases, it will not be true if one
     * of the threads dies prematurely...  pthreads has a habit of doing
     * this.  This should be fixed to use barrier_break when needed. */
    atomic_set(&req->linkage_counter, cf_header.num_threads);
    cr_barrier_init(&req->linkage_barrier, cf_header.num_threads);

#if HAVE_2_6_SIGNAL_STRUCT
    /* Check if signal struct is shared, so we know how to call clone() */
    retval = cr_kread(req->file, &req->shared_signal,
		      sizeof(req->shared_signal));
    if (retval < sizeof(req->shared_signal)) {
	retval = -EIO;
        goto out_free_req;
    }
#endif

#if HAVE_2_6_TASK_LINKAGE
    /* Now we reserve all the ids (pid, pgid, sid) we will need.
     * Note that this can get_free_page(), and thus no locks are held yet.
     */
    if (cr_reserve_ids(req)) {
        CR_WARN("%s [%d]:  PID conflict found by cr_reserve_ids()", __FUNCTION__, current->pid);
	retval = -EBUSY;
	goto out_free_req;	
    }
#endif

    /*
     * Check to see if the linkage is restorable
     */
    read_lock(&tasklist_lock);
    retval = cr_is_linkage_restorable(req);
    read_unlock(&tasklist_lock);
    if (!retval) {
        CR_WARN("%s [%d]:  PID conflict found by cr_is_linkage_restorable()", __FUNCTION__, current->pid);
        retval = -EBUSY;
	goto out_free_pids;
    }

    /* Normal case.  We don't want to ix the req, since we will need it
     * again */
    return 0;

out_free_pids:
#if HAVE_2_6_TASK_LINKAGE
    cr_release_ids(req);
#endif
out_free_req:
    release_rstrt_req(req);
    priv->rstrt_req = NULL;

    return retval;
}

// Check that a request is "done", from the point of view of the
// requestor.
// Returns >0 is completed, 0 if not completed, <0 on error
static
int check_done(cr_rstrt_req_t *req)
{
	/* Check that it is just me here,
	 * and that one or more children either started or died */
	return ((atomic_read(&req->ref_count) == 1) &&
		!test_bit(0, &req->need_children))	? 1 : 0;
}
	
// cr_rstrt_done(timeval)
//
// Test/wait for a restart request to complete (tasks are alive).
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// XXX: *should* return PID of "root" process (session leader)
int cr_rstrt_done(struct file *filp, void *arg)
{
    cr_pdata_t		*priv;
    cr_rstrt_req_t	*req;
    struct timeval	timeval;
	
    unsigned long timeout;
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    priv = filp->private_data;
    if (!priv) {
	return -EINVAL;
    }
    req = priv->rstrt_req;
    if (!req) {
	return -EINVAL;
    }

    // Fetch/decode the optional timeout from userspace
    timeout = MAX_SCHEDULE_TIMEOUT;
    if (arg) {
	retval = copy_from_user(&timeval, arg, sizeof(timeval));
	if (retval) {
	    return retval;
	}

	timeout = timeval_to_jiffies(&timeval);
    } 

    timeout = wait_event_interruptible_timeout(req->wait, (retval = check_done(req)), timeout);

    // Copy remaining time back to userspace if applicable
    if (arg) {
	jiffies_to_timeval(timeout, &timeval);
	copy_to_user(arg, &timeval, sizeof(timeval));
	// like select() we fail silently if user timeval is R/O
    }

    // Check if we were interrupted
    if (!retval && timeout) {
	retval = -EINTR;
    }

    return retval;
}

// cr_rstrt_reap()
//
// Reap a completed restart.
// Returns:
// 	req->result on success
// 	-EINVAL if no un-reaped restart request is associated w/ this fd
// 	-EAGAIN if the request is not completed
int
cr_rstrt_reap(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    priv = filp->private_data;
    if (priv) {
	req = priv->rstrt_req;
	if (req) {
	    retval = -EAGAIN;
	    if (check_done(req)) {
		priv->rstrt_req = NULL;
		retval = req->result;
		(void) release_rstrt_req(req);
	    }
	}
    }

    return retval;
}

/*
 * cr_linkage_find_by_pid
 *
 * Returns a pointer to a linkage entry with l.myptr == myptr
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_pid(cr_rstrt_req_t *req, 
                       int pid)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<req->linkage_size && entry == NULL; ++i) {
        if (req->linkage[i].pid == pid) {
	    entry = &(req->linkage[i]);
	}
    }

    return entry;
}

/*
 * cr_linkage_find_by_old_task
 *
 * Returns a pointer to a linkage entry with l.myptr == myptr
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_old_task(cr_rstrt_req_t *req, 
                            struct task_struct *myptr)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    if (myptr != NULL) {
        for (i=0; i<req->linkage_size; ++i) {
            if (req->linkage[i].myptr == myptr) {
	        entry = &(req->linkage[i]);
		break;
	    }
        }
    }

    return entry;
}

/*
 * cr_linkage_find_by_new_task
 *
 * Returns a pointer to a linkage entry with l.task == task
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_new_task(cr_rstrt_req_t *req,
                            struct task_struct *task)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<req->linkage_size; ++i) {
        if (req->linkage[i].tl_task == task) {
	    entry = &req->linkage[i];
	    break;
	}
    }

    return entry;
}

/*
 * cr_linkage_get_parent
 *
 * Figure out who your parent process ought to be
 *
 * This takes care of making sure that if your new parent is cr_restart, then
 * things are ok.  Also takes care of init.
 *
 * XXX:  Well, it will take care of init.  For right now, we just assume
 * that init_task is the cr_restart task.  so checkpointed tasks that used
 * to be children of init are now children of cr_restart.
 */
struct cr_context_tasklinkage *
cr_linkage_get_parent(cr_rstrt_req_t *req,
                      struct task_struct *old_pptr)
{
    struct cr_context_tasklinkage *parent_entry;

    parent_entry = cr_linkage_find_by_old_task(req, old_pptr);
    /* No entry for this process, we'll reattatch it as a new child
     * of cr_restart. */
    if (parent_entry == NULL) {
        parent_entry = &req->linkage[0];
    }

    /* XXX:  Kludge
     * We don't deal with children of init correctly yet.  This makes sure
     * that if your parent _WAS_ init at the time of checkpoint, that your
     * new parent will be cr_restart
     */
    if (parent_entry == &req->linkage[1]) {
        parent_entry = &req->linkage[0];
    }

    return parent_entry;
}

/*
 * cr_linkage_get_real_parent
 *
 * For now, just assume the same.  Can fix this later by just
 * duplicating the above code.
 */
struct cr_context_tasklinkage *
cr_linkage_get_real_parent(cr_rstrt_req_t *req,
                         struct task_struct *old_opptr)
{
    return cr_linkage_get_parent(req, old_opptr);
}

#if HAVE_2_6_TASK_LINKAGE
/*
 * Add task to all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_hash(struct task_struct *task, int is_leader)
{
    SET_LINKS(task);
    attach_pid(task, PIDTYPE_PID, task->pid);
#if HAVE_TASK_PIDS_PID
    attach_pid(task, PIDTYPE_TGID, task->tgid);
#endif
    if (is_leader) {
#if HAVE_TASK_PIDS_PID_LINK
        attach_pid(task, PIDTYPE_TGID, task->tgid);
#endif
	attach_pid(task, PIDTYPE_PGID, cr_task_pgrp(task));
	attach_pid(task, PIDTYPE_SID, cr_task_session(task));
    } else {
#if HAVE_TASK_PIDS_PID_LINK
        link_pid(task, task->pids + PIDTYPE_TGID,
		 &task->group_leader->pids[PIDTYPE_TGID].pid);
#endif
    }
}

/*
 * Remove task from all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_unhash(struct task_struct *task)
{
    detach_pid(task, PIDTYPE_PID);
    detach_pid(task, PIDTYPE_TGID);
    if (thread_group_leader(task)) {
	detach_pid(task, PIDTYPE_PGID);
	detach_pid(task, PIDTYPE_SID);
    }
    REMOVE_LINKS(task);
}
#endif


/* 
 * cr_do_restore_linkage
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every process contained in the restart request.  All other threads
 * should be blocked and prevented from returning to user space before entry
 * into this function.
 *
 * NOTE:  Must be called with tasklist_lock held.
 * Watch it inside of here, since you're holding a spinlock, you can't
 * do very much, and you can't waste too much time.  Just be concise and
 * to the point.
 *
 */
static int
cr_do_restore_linkage(cr_rstrt_req_t *req)
{
    int retval = 0;
    int i;
    struct cr_context_tasklinkage *link;

    /* XXX:  thread_group has to be handled specially, since we may only
     * construct this structure by fiddling with the task_struct directly.
     *
     * XXX:  thread_group is not restored
     */

    /*
     * Make a pass through to initialize some fields and to
     * translate old task pointers to new ones
     */
    for (i=0; i<req->linkage_size; ++i) {
        link = &req->linkage[i];
	if (link->tl_task == NULL) {
            /* Means that we have an entry for which there is no
	     * corresponding task! */

	    /* XXX:  This can actually happen in normal cases if a thread
	     * exits...  Ignore for now...  Can probably pull this code
	     * out later. */
	    CR_ERR("%s [%d]:  ERROR:  No task for this PID. %d",
		     __FUNCTION__, current->pid, link->pid);
	    retval = -EINVAL;
	    goto out;
	}

	if (i < 2) {
	    /* Special case cr_restast task and init */
	    continue;
        }

	/* translate the linkage pointers from saved to restored */
	link->real_parent = cr_linkage_get_real_parent(req, link->real_parent)->tl_task;
        link->parent = cr_linkage_get_parent(req, link->parent)->tl_task;
#if HAVE_2_4_TASK_LINKAGE
	INIT_LIST_HEAD(&link->thread_group);
#elif HAVE_2_6_TASK_LINKAGE
	link->group_leader = cr_linkage_find_by_old_task(req, link->group_leader)->tl_task;
#endif
    }

    /* Unlike the last pass, this pass is decisive.  If it doesn't work here,
     * then we _have_ to abort!  (or skip) */
    if (!cr_is_linkage_restorable(req)) {
        retval = -EBUSY;
        goto out;	
    }

    /*
     * init is special cased.  For now, just hack the data structure
     * so that references to the old init become references to the
     * cr_restart task
     */

    /* OK, now that the pointers have been calculated, we can go through
     * and actually perform the swap */
#if HAVE_2_4_TASK_LINKAGE
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task;

        link = &req->linkage[i];
	task = link->tl_task;

	/* Remove existing links with parent, siblings, hashes, etc. */
	unhash_pid(task);
	REMOVE_LINKS(task);

        /* Set simple links to parent(s), aunts, uncles, etc. */
        task->p_pptr  = link->parent;
	task->p_opptr = link->real_parent;

	/* next set the various ids */
	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->exit_signal = link->exit_signal;

	/* Now relink, rehash for the new relatives and ids */
	SET_LINKS(task);
	hash_pid(task);

	// XXX:  Do thread_next

	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }
#elif HAVE_2_6_TASK_LINKAGE
    /* Keeping the pid hashes/lists correct requires multiple passes */

    /* Pass 1: unhash and unlink all the tasks */
    for (i=2; i<req->linkage_size; ++i) {
	/* Remove existing links with parent, siblings, hashes, etc. */
	struct task_struct *task = req->linkage[i].tl_task;
	cr_do_unhash(task);
    }

    /* Pass 2: change all the simple links and ids */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task;

        link = &req->linkage[i];
	task = link->tl_task;

        /* Set simple links to (step-)parent and task group leader */
        task->parent       = link->parent;
	task->real_parent  = link->real_parent;
	task->group_leader = link->group_leader;

	/* next set the various ids */
	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->exit_signal = link->exit_signal;

	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }

    /* Pass 3: relink/rehash only thread group leaders
     * These must exist before their followers can be linked in. */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].tl_task;

        if (thread_group_leader(task)) {
		cr_do_hash(task, 1);
		/* To look like we called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }

    /* Pass 4: relink/rehash non-leaders */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].tl_task;

        if (!thread_group_leader(task)) {
		cr_do_hash(task, 0);
		/* To look like we were fork()ed by leader, who called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id + 1;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }
#endif

    /* XXX A hack to return a "lead" pid. */
    for (i=2; i<req->linkage_size; ++i) {
	struct task_struct *task = req->linkage[i].tl_task;

        if (CR_PARENT(task) == req->cr_restart_task) {
	    if (!req->result) { /* ensure we don't clobber an error indication */
		req->result = task->pid;
	    }
	    break;
	}
    }

    /* on success retval is zero when we get here */

out:
    return retval;
}

/*
 * cr_restore_linkage
 *
 * This function implements a barrier, and hence must be called by all
 * restarting tasks before their return to user space.  Once the barrier
 * condition is reached (tasks entered == tasks created) the last thread
 * to enter restores the linkage.  At that point, the other tasks are
 * woken, and all tasks all allowed to continue
 *
 * This grabs tasklist_lock.
 *
 * Use cr_load_linkage to place the linkage structure into memory before
 * this function is called.
 *
 * Two _magic_ values in linkage structure:
 *   linkage[0] corresponds to the _original_ cr_restart task
 *     XXX:  Should this change to the initial baby?  -- sessions?
 *   linkage[1] corresponds to the init task
 */
static int
cr_restore_linkage(cr_rstrt_req_t *req, int pid)
{
    int retval;
    int my_old_pid;
    struct cr_context_tasklinkage *my_entry = NULL;

    /* Save this for debugging */
    my_old_pid = current->pid;

    /* Look up your entry in the old linkage */
    if (pid > 0) {
        my_entry = cr_linkage_find_by_pid(req, pid);
    } else {
        CR_WARN("Called after vmadump failure.");
    }

    /* Associate your old PID with your task in the linkage */
    if (my_entry == NULL) {
        CR_ERR("%s [%d]: ERROR:  No linkage entry for PID %d",
	       __FUNCTION__, current->pid, pid);

	/*
	 * NOTE:  We DO NOT ABORT here.  Instead, we wait for the linkage to be restored,
	 * and rely on the cr_do_restore_linkage code to notice the problem.
	 */
    } else {
	get_task_struct(current);
        my_entry->tl_task = current;
    }

    /* First, we need to stop, so that none of the children start to
     * execute until their PIDs have been restored. */

    if (atomic_dec_and_test(&req->linkage_counter)) {
        /* True only for the last task.  All other tasks have also reached
	 * this point and will block at the barrier below while we do
	 * the remaining work. */

        CR_KTRACE_HIGH_LVL("Now restoring the linkage... ");
        /* Now grab tasklist_lock while we futz with everyones pointers */
        write_lock(&tasklist_lock);
	
	/* Do it! */
        retval = cr_do_restore_linkage(req);

	/* Let the rest of the system come back */
	write_unlock(&tasklist_lock);

	if (retval < 0) {
            CR_KTRACE_HIGH_LVL("Linkage restore failed.  Releasing reserved ids.");
#if HAVE_2_6_TASK_LINKAGE
            cr_release_ids(req);
#endif
	} else {
            CR_KTRACE_HIGH_LVL("Linkage restore finished...");
	}

	retval = 0;
    } else {
	/* Nothing to do, so just return successful */
	retval = 0;
    }

    /* Now wait until everything quiets down */
    cr_barrier_enter(&req->linkage_barrier);

    /* XXX: check for broken barrier */

    write_lock(&tasklist_lock);
    /* placing a NULL entry in the linkage table forces cr_do_restore_linkage to abort. */
    my_entry->tl_task = NULL;
    put_task_struct(current);
    write_unlock(&tasklist_lock);

    return retval;
}

/*
 * cr_fgets
 *
 * Sort of like fgets.
 *   Returns >= 0 on success.
 *   0 return means string is NULL.
 *   Returns number of bytes read in bytesread.
 */
int
cr_fgets(char *buf, int size, struct file *filp, int *bytesread)
{
    int ret;
    int read;
    int len = 0;

    ret = -EINVAL;
    if (!(bytesread)) {
        CR_ERR("cr_fgets: NULL bytesread passed as argument!");
        goto out;        
    }
    read = 0;

    ret = -EINVAL;
    if (!(buf)) {
        CR_ERR("cr_fgets: NULL buffer pass as argument!");
        goto out_bytesread;
    }

    ret = -EINVAL;
    if ((size < 0) || (size > PAGE_SIZE)) {
        CR_ERR("cr_fgets:  Bad buffer size.");
        goto out;
    }

    ret = cr_kread(filp, &len, sizeof(len));
    if (ret < 0) {
        CR_ERR("cr_fgets: Bad file read.");
        goto out_bytesread;
    } else if (ret != sizeof(len)) {
        read = ret;
        ret = -EIO;
        CR_ERR("cr_fgets: Short read.");
        goto out_bytesread;
    }
    read = ret;

    if (len == 0) {
        /* old string was NULL.  i.e. never initialized rather than a
	 * pointer to an empty string */
        ret = 0;
        goto out_bytesread;
    } else if ((len < 0) || (len > PAGE_SIZE) || (len > size)) {
        ret = -EINVAL;
        CR_ERR("cr_fgets:  Bad string length %d in file.", len);
        goto out_bytesread;
    } 

    ret = cr_kread(filp, buf, len);
    if (ret < 0) {
        CR_ERR("cr_fgets:  Bad file read.");
        goto out;
    } else if (ret != len) {
        read += ret;
        ret = -EIO;
        CR_ERR("cr_fgets:  Short read.");
        goto out_bytesread;
    }
    read += ret;

out_bytesread:
    *bytesread = read;
out:
    return ret;
}

/*
 * From Linux 2.4.20 fs/file.c
 */
fd_set * cr_alloc_fdset(int num)
{
        fd_set *new_fdset;
        int size = num / 8;

        if (size <= PAGE_SIZE)
                new_fdset = (fd_set *) kmalloc(size, GFP_KERNEL);
        else
                new_fdset = (fd_set *) vmalloc(size);
        return new_fdset;
}

/*
 * From Linux 2.4.20 fs/file.c
 */
void cr_free_fdset(fd_set *array, int num)
{
        int size = num / 8;

        if (!array) {
                printk(KERN_ERR "%s array = 0 (num = %d)\n",
                                __FUNCTION__, num);
                return;
        }

        if (num <= __FD_SETSIZE) /* Don't free an embedded fdset */
                return;
        else if (size <= PAGE_SIZE)
                kfree(array);
        else
                vfree(array);
}

/*
 * cr_restore_files_max_fdset
 *
 * Restores our fd_array and fdset_array to correct values to reopen old files.
 *
 * Adapted from expand_fdset in fs/file.c from Linux 2.4.20
 */
static int
cr_restore_files_max_fdset(struct files_struct *files, int nr)
{
    fd_set *new_cloexec_fds = NULL;
    fd_set *new_open_fds = NULL;
    int retval;

    /* check system limits */
    retval = -EMFILE;
    if (files->max_fds >= NR_OPEN || nr >= NR_OPEN)
        goto out;

    retval = 0;
    cr_file_lock_write(&files->file_lock);
    if (nr <= files->max_fdset) {
	cr_file_lock_unwrite(&files->file_lock);
	goto out;
    }

    retval = -ENOMEM;
    cr_file_lock_unwrite(&files->file_lock);
    new_cloexec_fds = cr_alloc_fdset(nr);
    new_open_fds = cr_alloc_fdset(nr);
    if (!new_open_fds || !new_cloexec_fds) {
	CR_ERR("Unable to allocate fdset.");
	goto out;
    }

    retval = 0;
    cr_file_lock_write(&files->file_lock);
    /* Copy the existing tables and install the new pointers */
    if (nr > files->max_fdset) {
	int i = files->max_fdset / (sizeof(unsigned long) * 8);
	int count = (nr - files->max_fdset) / 8;

	/* 
	 * Don't copy the entire array if the current fdset is not yet initialised.  
	 */
	if (i) {
	    memcpy(new_open_fds, files->open_fds, files->max_fdset / 8);
	    memcpy(new_cloexec_fds, files->close_on_exec, files->max_fdset / 8);
	    memset(&new_open_fds->fds_bits[i], 0, count);
	    memset(&new_cloexec_fds->fds_bits[i], 0, count);
	}

	nr = xchg(&files->max_fdset, nr);
	new_open_fds = xchg(&files->open_fds, new_open_fds);
	new_cloexec_fds = xchg(&files->close_on_exec, new_cloexec_fds);
	cr_file_lock_unwrite(&files->file_lock);
	cr_free_fdset(new_open_fds, nr);
	cr_free_fdset(new_cloexec_fds, nr);
	return 0;
    }
    /* Somebody expanded the array while we slept ... */

    cr_file_lock_unwrite(&files->file_lock);

out:
    if (new_open_fds)
	cr_free_fdset(new_open_fds, nr);
    if (new_cloexec_fds)
	cr_free_fdset(new_cloexec_fds, nr);
    return retval;
}


/*
 * frees an fd_array.
 *
 * Borrowed from Linux 2.4.20 fs/file.c
 */
void cr_free_fd_array(struct file **array, int num)
{
        int size = num * sizeof(struct file *);

        if (!array) {
                printk(KERN_ERR "%s array = 0 (num = %d)\n",
                                __FUNCTION__, num);
                return;
        }

        if (num <= NR_OPEN_DEFAULT) /* Don't free the embedded fd array! */
                return;
        else if (size <= PAGE_SIZE)
                kfree(array);
        else
                vfree(array);
}

/*
 * Restores the fd array to the proper size.
 *
 * Again, borrowed heavily from linux/fs/file.c
 */
static int
cr_restore_files_max_fds(struct files_struct *files, int nr)
{
    int size = nr * sizeof(struct file *);
    int retval;
    struct file **new_fd_array;

    cr_file_lock_write(&files->file_lock);
    if (nr <= files->max_fds) {
	cr_file_lock_unwrite(&files->file_lock);
	return 0;
    }

    retval = -ENOMEM;
    cr_file_lock_unwrite(&files->file_lock);

    /* allocate new array */
    if (size <= PAGE_SIZE)
	new_fd_array = (struct file **) kmalloc(size, GFP_KERNEL);
    else
	new_fd_array = (struct file **) vmalloc(size);

    if (!new_fd_array)
	goto out;

    /* now copy the stuff in the old array into the new one. */
    cr_file_lock_write(&files->file_lock);
    if (nr > files->max_fds) {
	struct file **old_fds;
	int i;

	old_fds = xchg(&files->fd, new_fd_array);
	i = xchg(&files->max_fds, nr);

	/* Don't copy/clear the array if we are creating a new fd array for fork() */
	if (i) {
	    memcpy(new_fd_array, old_fds, i * sizeof(struct file *));
	    /* clear the remainder of the array */
	    memset(&new_fd_array[i], 0, (nr - i) * sizeof(struct file *));

	    cr_file_lock_unwrite(&files->file_lock);
	    cr_free_fd_array(old_fds, i);
	}
    } else {
	/* Somebody expanded the array while we slept ... */
	cr_file_lock_unwrite(&files->file_lock);
	if (size <= PAGE_SIZE)
            kfree(new_fd_array);
        else
            vfree(new_fd_array);
    }
    retval = 0;

out:
    return retval;
}

/*
 * cr_restore_files_struct
 *
 * Restore the files_struct (file table).
 *
 * This function just expands our fd array to the correct size.
 *
 * In the future, if we need any information that applies to all opened files
 * that info can be placed here.
 */
static int
cr_restore_files_struct(cr_rstrt_req_t *req)
{
    struct cr_files_struct cr_fs;
    struct file *c_filp;
    int retval = 0;

    c_filp = cr_src_gett(req, cr_files_obj);
    if (!c_filp) {
	CR_ERR("Unable to find context file to load cr_files_struct.");
	goto out_no_cfilp;
    }

    retval = cr_kread(c_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
	CR_ERR("cr_fs: Bad file read!");
	retval = -EIO;
	goto out_putt;
    }

    retval = 0;
    if (cr_fs.cr_obj != cr_files_obj) {
	CR_ERR("cr_fs: Invalid cr_files_struct!");
	retval = -EINVAL;
	goto out_putt;
    }

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (((unsigned long) cr_fs.cr_max_fdset) > CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur) {
        CR_KTRACE_HIGH_LVL("Too many files %d > %lu", cr_fs.cr_max_fdset, CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur);
	goto out_putt;
    }

    retval = cr_restore_files_max_fdset(current->files, cr_fs.cr_max_fdset);
    if (retval < 0) {
	CR_ERR("Failed to resize the fdset array. (err=%d)", retval);
	goto out_putt;
    }

    retval = cr_restore_files_max_fds(current->files, cr_fs.cr_max_fds);
    if (retval < 0) {
	CR_ERR("Failed to resize the fd array. (err=%d)", retval);
	goto out_putt;
    }

    cr_file_lock_write(&current->files->file_lock);
    if (cr_fs.cr_next_fd > current->files->max_fdset) {
        /* shouldn't ever happen */
        retval = -EINVAL;
        cr_file_lock_unwrite(&current->files->file_lock);
	CR_ERR("Failed to restore next_fd.");
	goto out_putt;
    }

// #error make sure this fd is closed?
    // current->files->next_fd = cr_fs.cr_next_fd;
    cr_file_lock_unwrite(&current->files->file_lock);

    retval = 0;

out_putt:
    cr_src_putt(c_filp);
out_no_cfilp:
    return retval;
}

/*
 * cr_load_file_info
 * 
 * Reads in a file_info from disk, returns it.
 */
static int
cr_load_file_info(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    struct file *filp;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    filp = cr_src_gett(req, cr_file_obj);
    if (!filp) {
        retval = -EBADF;
        goto out;
    }

    retval = cr_kread(filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
        CR_ERR("cr_load_file_info: Bad read! retval=%d", retval);
        retval = -EIO;
	goto out_putt;
    }

    retval = -EINVAL;
    if (file_info->cr_type != cr_file_info_obj) {
        CR_ERR("cr_load_file_info: Garbage in context file! (type=%d)", file_info->cr_type);
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(filp);
out:
    return retval;
}

/*
 * cr_fd_claim
 *
 * Atomically checks and claims an fd
 *
 * returns < 0 on conflict w/ an existing fd
 */
static int
cr_fd_claim(int fd)
{
    int retval;

    /* Mark the fd in use */
    cr_file_lock_write(&current->files->file_lock);
    if (FD_ISSET(fd, current->files->open_fds)) {
	retval = -EBUSY;
    } else {
	retval = 0;
	FD_SET(fd, current->files->open_fds);
	FD_CLR(fd, current->files->close_on_exec);
    }
    cr_file_lock_unwrite(&current->files->file_lock);

    return retval;
}

/*
 * cr_open_fd: open a file - accepts file descriptor as argument
 * 
 * NOTE:  Does not attempt to expand the fdset array. i.e. aborts if fd > files->max_fdset
 * Use cr_restore_files_struct to expand fdset to the correct size.
 *
 * NOTE:  Does not check limits on number of open files.
 */ 
static int
cr_open_fd(int fd, char *path, int flags, int mode)
{
    int retval;
    struct file *filp;

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (fd >= CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur)
	goto out;

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (fd >= current->files->max_fds)
	goto out;

    /* Mark the fd in use */
    retval = cr_fd_claim(fd);
    /* make sure it's not in use already.  shouldn't ever happen */
    if (retval) {
        CR_WARN("File opened unexpectedly.");
        goto out;
    }

    /* open the file */
    filp = filp_open(path, flags, mode);
    if (IS_ERR(filp)) {
        retval = PTR_ERR(filp);
        goto out;
    }

    /* install the filp in the fd array -- note race see fd_install */
    fd_install(fd, filp);

    retval=0;

out:
    return retval;
}

/*
 * cr_anonymous_rename
 *
 * Rewrite a filename to an anonymous value
 *
 * len is strlen and size is buffer size
 */
static char *
cr_anonymous_rename(char *in_buf, unsigned long id)
{
    size_t len;
    char *out_buf, *p;

    len = strlen(in_buf);
    if (len >= PAGE_SIZE) {
	/* XXX: probably not what we want to do */
	len = PAGE_SIZE - 1;
    }

    out_buf = (char *) __get_free_page(GFP_KERNEL);
    if (out_buf == NULL) {
	goto out;
    }
    memcpy(out_buf, in_buf, len+1);

    p = out_buf + len;
    while ((p != out_buf) && (*p != '/')) {
	--p;
    }
    ++p;
    len = p - out_buf;

#if BITS_PER_LONG == 32
    if (len > PAGE_SIZE - 20) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%08lx", (unsigned int)current->pid, id);
#elif BITS_PER_LONG == 64
    if (len > PAGE_SIZE - 24) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%016lx", (unsigned int)current->pid, id);
#else
    #error "No value for BITS_PER_LONG"
#endif

    return out_buf;

out_free:
    free_page((unsigned long)out_buf);
out:
    return NULL;
}

/*
 * cr_restore_open_file
 *
 * reopens files
 */
static int
cr_restore_open_file(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    char *buf;
    struct file *c_filp;
    struct cr_open_file open_file;
    int len;

    CR_KTRACE_FUNC_ENTRY();

    c_filp = cr_src_gett(req, cr_file_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        CR_ERR("Couldn't allocate buffer for file name.");
        retval = -ENOMEM;
        goto out_putt;
    }
    
    /* read in the cr_open_file struct */
    retval = cr_kread(c_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    if (open_file.cr_type != cr_open_file_obj) {
        CR_ERR("cr_restore_open_file.  Garbage in context file!");
    }

    /* now read out the name */
    retval = cr_fgets(buf, PAGE_SIZE, c_filp, &len);
    if (retval < 0) {
        CR_ERR("cr_restore_open_file - pwd Bad read!");
        retval = -EIO;
        goto out_free;
    } else if (retval == 0) {
        /* 0 corresponds to NULL, should never have happened. */
        CR_ERR("A file with no name?");
	retval = -EINVAL;
	goto out_free;
    }

    if (do_not_restore) {
        retval = 0;
	goto out_free;
    }

    /* reopen file */
    retval = cr_open_fd(file_info->fd, buf, open_file.f_flags, open_file.i_mode);
    if(retval) {
        goto out_free;
    }

    /* restore position in file */
    retval = sys_lseek(file_info->fd, open_file.f_pos, 0);
    if (retval < 0) {
        CR_ERR("Couldn't restore file pointer.");
	goto out_free;
    }

    /* restore size of file */
    if (open_file.f_flags & FMODE_WRITE) {
        retval = sys_ftruncate(file_info->fd, open_file.i_size);
        if (retval < 0) {
            CR_ERR("Couldn't restore file length.");
	    goto out_free;
        }
    }

    retval = 0;

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_dir(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_link(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

/*
 * f_flags - same as in user space
 * f_mode  - funny kernel thing (can reconstruct from f_flags)
 *   00 - no permissions needed
 *   01 - read permission needed
 *   10 - write permission needed
 *   11 - read/write permissions needed
 */
static int
cr_posix_flags_to_f_mode(int posix_flags)
{
    return ((posix_flags+1) & O_ACCMODE);

}

/*
 * This is more or less the same thing as open(), with the major exception
 * that we DO NOT BLOCK.  (Normal open() does block in certain cases.)
 * POSIX tells us that if we open write-only when no readers are present
 * we must either block, or fail w/ errno=ENXIO if O_NONBLOCK is set.
 * Therefore, we can't JUST set O_NONBLOCK.  We must ALSO mess w/ the
 * flags to open read/write when the user asks for write-only.
 *
 * XXX: this could easily break for a client attaching to a FIFO for
 * which it has write permission but not read permission.  We should
 * really just see about reproducing more of the fifo_open instead.
 */
int
cr_open_named_fifo(int fd, struct dentry *dentry, struct vfsmount *mnt, int flags)
{
    int retval = 0;
    struct file *filp;
    int tmp_flags;

    tmp_flags = flags | O_NONBLOCK;
    if (flags & O_WRONLY) {
        /* We promote W-only to RW, dropping the R later */
	tmp_flags = (tmp_flags & ~O_ACCMODE) | O_RDWR;
    }

    filp = dentry_open(dget(dentry), mntget(mnt), tmp_flags);
    if (IS_ERR(filp)) {
	retval = PTR_ERR(filp);
        goto out;
        /* get out */
    }

    /* Check here that we actually got a FIFO */
   if (!S_ISFIFO(filp->f_dentry->d_inode->i_mode)) {
	retval = -EEXIST;
        CR_ERR("encountered non-FIFO with same name as desired FIFO.");
        goto out;
   }
	
    /* Fixup: */
    down(PIPE_SEM(*filp->f_dentry->d_inode));
    filp->f_flags = flags;
    if (flags & O_WRONLY) {
        /* Fix up the mode and the inode if we faked RDWR */
        filp->f_mode = cr_posix_flags_to_f_mode(O_WRONLY);
	filp->f_op = &write_fifo_fops;
	PIPE_READERS(*filp->f_dentry->d_inode)--;
	wake_up_interruptible(PIPE_WAIT(*filp->f_dentry->d_inode));
    }
    up(PIPE_SEM(*filp->f_dentry->d_inode));

    /* Install it */
    retval = cr_fd_claim(fd);
    if (retval < 0) {
        CR_ERR("File descriptor %d in use.", fd);
	goto out;
    }
    fd_install(fd, filp);

out:
    return retval;
}

/*
 *
 */

/* half of do_pipe - set up new file pointer to an existing pipe */
static int
cr_open_unnamed_pipe(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	struct cr_fifo *cf_fifo)
{
    int retval = -EINVAL;
    struct file *filp;

    CR_KTRACE_LOW_LVL("Opening FIFO ID %p", cf_fifo->fifo_id);

    /* unnamed pipe */
    filp = get_empty_filp();

    if (!filp) {
        retval = -ENOMEM;
	goto out_err;
    }

    /* first the boring stuff */
    filp->f_pos     = cf_fifo->fifo_pos;
    filp->f_flags   = cf_fifo->fifo_flags;
    filp->f_mode    = cr_posix_flags_to_f_mode(filp->f_flags);
    filp->f_version = cf_fifo->fifo_version;

    /* now the dentry and vfsmnt, to attach us to the filesystem */
    cr_find_object(req->map, cf_fifo->fifo_dentry, (void **) &filp->f_dentry);
    if (!(filp->f_dentry)) {
	CR_ERR("cr_open_unnamed_pipe: Couldn't look up dentry %p in object map", cf_fifo->fifo_dentry);
        goto out_free;
    }

    cr_find_object(req->map, cf_fifo->fifo_vfsmnt, (void **) &filp->f_vfsmnt);
    if (!(filp->f_vfsmnt)) {
	CR_ERR("cr_open_unnamed_pipe: Couldn't look up vfsmnt %p in object map", cf_fifo->fifo_vfsmnt);
        goto out_free;
    }

    filp->f_dentry = dget(filp->f_dentry);
    if (!filp->f_dentry) {
	CR_ERR("dget(%p) failed", filp->f_dentry);
        goto out_free;
    }
        
    filp->f_vfsmnt = mntget(filp->f_vfsmnt);
    if (!filp->f_vfsmnt) {
	CR_ERR("mntget(%p) failed", filp->f_vfsmnt);
        goto out_dput;
    }

    /* now the file_operations f_op */
    switch(filp->f_mode) {
        case 1:
	  /* rdonly */
	  filp->f_op = &read_pipe_fops;
          PIPE_READERS(*filp->f_dentry->d_inode)++;
	  break;
	case 2:
	  /* wronly */
	  filp->f_op = &write_pipe_fops;
          PIPE_WRITERS(*filp->f_dentry->d_inode)++;
	  break;
	case 3:
	  /* rdwr */
	  filp->f_op = &rdwr_pipe_fops;
	  if (filp->f_mode & FMODE_READ) 
	    PIPE_READERS(*filp->f_dentry->d_inode)++;
          if (filp->f_mode & FMODE_WRITE) 
	    PIPE_WRITERS(*filp->f_dentry->d_inode)++;
	  break;
	default:
	  CR_ERR("Bad access mode for pipe.");
	  retval = -EBADF;
	  goto out_mntput;
	  break;
    }

    /* allocate the fd */
    retval = cr_fd_claim(file_info->fd);
    if (retval) {
        /* this should never happen.  Something else is already using our fd.  
	 * Race? Or maybe forgot to set do_not_restore... */
        CR_ERR("File descriptor %d in use.", file_info->fd);
        goto out_mntput;
    }

    /* install the fd */
    fd_install(file_info->fd, filp);

    return 0;

    /* only get here if something went wrong. */
out_mntput:
    mntput(filp->f_vfsmnt);
out_dput:
    dput(filp->f_dentry);
out_free:
    put_filp(filp);
out_err:
    return retval;
}

/* Use path_lookup() if present as a function or a macro */
#if defined(path_lookup) || defined(CR_KCODE_path_lookup) || defined(CR_EXPORTED_KCODE_path_lookup)
  /* Macro or function (possibly unexported) */
  #define cr_path_lookup path_lookup
#else
  /* based on linux/fs/namei.c:path_lookup in later 2.4.x kernels */
  static int cr_path_lookup(const char *path, unsigned flags, struct nameidata *nd)
  {
    int retval = 0;
    if (path_init(path, flags, nd)) {
      retval = path_walk(path, nd);
    }
    return retval;
  }
#endif

/* from on linux/fs/namei.c:lookup_create */
/* Differs in that the balancing up() takes place on error */
static struct dentry *cr_lookup_create(struct nameidata *nd, int is_dir)
{
    struct dentry *dentry;

    down(&nd->dentry->d_inode->i_sem);
    dentry = ERR_PTR(-EEXIST);
    if (nd->last_type != LAST_NORM) {
	goto out_fail;
    }
    dentry = lookup_hash(&nd->last, nd->dentry);
    if (IS_ERR(dentry)) {
	goto out_fail;
    }
    if (!is_dir && nd->last.name[nd->last.len] && !dentry->d_inode) {
	goto out_noent;
    }
    return dentry;

out_noent:
    dput(dentry);
    dentry = ERR_PTR(-ENOENT);
out_fail:
    up(&nd->dentry->d_inode->i_sem);
    return dentry;
}

/* cr_mknod - based on linux/fs/namei.c:sys_mknod
 *
 * Creates regular files, fifos or devices, making them anonymous (unlinked)
 * if desired.
 * Returns a dentry for the resulting filesystem objects, and the corresponding
 * vfsmnt can be obtained in nd->mnt.  Together these two can be passed
 * to dentry_open(), even for an unlinked inode.
 * In the event of an error, no dput() or path_release() is required,
 * otherwise they are.
 * In the event that an object exists with the given name, it will be
 * returned.  Therefore the caller must check the type of the resulting
 * dentry (addition of a 'uniq' flag might change this behavior later).
 * It is also possible to get -EEXIST from the lookup_create step.
 */
static struct dentry *
cr_mknod(struct nameidata *nd, char *name, int mode, int unlinked_id)
{
    struct dentry * dentry;
    char *_name;
    int err;

    if (unlinked_id) {
	/* Generate a replacement name which we will use instead of the original one. */
	_name = cr_anonymous_rename(name, unlinked_id);
	if (!_name) {
	    CR_ERR("cr_mknod - failed to rename unlinked object");
	    err = -ENOMEM;
	    goto out;
	}
    } else {
	_name = name;
    }

    err = cr_path_lookup(_name, LOOKUP_PARENT, nd);
    if (err) {
	CR_ERR("Couldn't path_lookup for mknod %s.  err=%d.", name, err);
	goto out_free;
    }

    dentry = cr_lookup_create(nd, 0);
    if (IS_ERR(dentry)) {
	err = PTR_ERR(dentry);
	CR_ERR("Couldn't lookup_create for mknod %s.  err=%d.", name, err);
	goto out_release;
    }

    err = vfs_mknod(nd->dentry->d_inode, dentry, mode, 0 /* ignored */);
    if (unlinked_id && !err) { /* Note that we don't unlink if we failed to create */
	dget(dentry);	/* ensure unlink doesn't destroy the dentry */
	/* Note possibility of silent failure here: */
	(void)vfs_unlink(nd->dentry->d_inode, dentry);
	dput(dentry);
    }
    up(&nd->dentry->d_inode->i_sem);
    if (err == -EEXIST) {
	/* We fall through and return the dentry */
    } else if (err) {
	CR_ERR("Couldn't vfs_mknod %s.  err=%d.", name, err);
	goto out_put;
    }

    return dentry;

out_put:
    dput(dentry);
out_release:
    path_release(nd);
out_free:
    if (unlinked_id) {
	free_page((unsigned long)_name);
    }
out:
    return (struct dentry *)ERR_PTR(err);
}

static int
cr_creat_named_fifo(int fd, char *name, int flags, int mode, int unlinked)
{
    int retval;
    struct nameidata nd;
    struct dentry * dentry;

    /* make a new FIFO, or find the existing one */
    dentry = cr_mknod(&nd, name, mode, unlinked ? fd+1 : 0);
    if (IS_ERR(dentry)) {
	retval = PTR_ERR(dentry);
	CR_ERR("Failed to create named fifo %s.  err=%d.", name, retval);
	goto out;
    }

    /* open it */
    /* This will check for correct object type in the event of an existing object */
    retval = cr_open_named_fifo(fd, dentry, nd.mnt, flags);
    if (retval < 0) {
	CR_ERR("Couldn't open %s.  err=%d.", name, retval);
	goto out_dput;
    }

    retval = 0;

out_dput:
    dput(dentry);
    path_release(&nd);
out:
    return retval;
}

static int
cr_restore_pipe_buf(struct file *cf_filp, struct inode *p_inode, int buf_len, int do_not_restore_flag)
{
    int retval;
    void *buf;

    buf = (void *)kmalloc(buf_len, GFP_KERNEL);
    if (!buf) {
	retval = -ENOMEM;
	goto out;
    }

    retval = cr_kread(cf_filp, buf, buf_len);
    if (retval != buf_len) {
	CR_ERR("cr_restore_pipe_buf - Bad read.");
	retval = -EIO;
	goto out_free;
    }

    if (do_not_restore_flag) {
        CR_KTRACE_LOW_LVL("Unrestored pipe buffer %p length = %d", p_inode, buf_len);
        retval = 0;
        goto out_free;
    }

    if (down_interruptible(PIPE_SEM(*p_inode))) {
        retval = -ERESTARTSYS;
        goto out_free;
    }

#if HAVE_PIPE_INODE_INFO_BASE
    if (PIPE_LEN(*p_inode)) {
        /* someone put data in here already -- bail */
        retval = -EBUSY;
	goto out_up;
    }

    if (buf_len) {
	memcpy((void *)PIPE_BASE(*p_inode), buf, buf_len);
        PIPE_LEN(*p_inode) = buf_len;
        CR_KTRACE_LOW_LVL("Pipe buffer %p length = %d inode = %lu", p_inode, PIPE_LEN(*p_inode), p_inode->i_ino);
	retval = 0;
    }
#else
    {   struct pipe_inode_info *info = p_inode->i_pipe;
	char * p;
        int nrbufs;

	if (info->nrbufs != 0) {
            /* someone put data in here already -- bail */
            retval = -EBUSY;
	    goto out_up;
	}

	nrbufs = 0;
	p = buf;
	while (buf_len) {
	    size_t count = buf_len;
	    struct pipe_buffer *buf = info->bufs + nrbufs;
	    struct page *page = alloc_page(GFP_HIGHUSER);
	    if (count > PAGE_SIZE)
		count = PAGE_SIZE;
	    memcpy(kmap(page), p, count);
	    kunmap(page);
	    buf->page = page;
	    buf->offset = 0;
	    buf->len = count;
	    buf->ops = &anon_pipe_buf_ops;
	    ++nrbufs;
	    buf_len -= count;
	    p += count;
	}
	info->curbuf = 0;
	info->nrbufs = nrbufs;
    }
#endif

out_up:
    up(PIPE_SEM(*p_inode));
out_free:
    kfree(buf);
out:
    return retval;
}

/* 
 * pipe was instantiated already, so we connect our file descriptor
*  to the pipe inode.
*/
static int
cr_couple_fifo(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	struct cr_fifo *cf_fifo, char *name, int do_not_restore_flag)
{
    int retval;

    retval = -EINVAL;
    if (!cr_find_object(req->map, cf_fifo->fifo_id, NULL)) {
	CR_ERR("cr_couple_fifo - unable to find previously restored fifo");
	goto out;
    }

    CR_KTRACE_LOW_LVL("%s:  phase 2: connecting second end.", (char *) name);
    /* no obvious way to distinguish named pipes from unnamed besides 
     * filename.  anything with a '/' at the beginning is a named fifo, 
     * and anything with a 'p' is a pipe. */
    if (*((char *) name) == '/') {
        /* named fifo */
        if (do_not_restore_flag) {
            retval = 0;
        } else {
	    struct dentry *dentry;
	    struct vfsmount *mnt;
	    cr_find_object(req->map, cf_fifo->fifo_dentry, (void **) &dentry);
	    cr_find_object(req->map, cf_fifo->fifo_vfsmnt, (void **) &mnt);
	    retval = cr_open_named_fifo(file_info->fd, dentry, mnt, cf_fifo->fifo_flags);
        }
        if (retval < 0) {
            CR_ERR("cr_couple_fifo:  unable to open named fifo: %s",
                   (char *) name);
            goto out;
        }
    } else if (*((char *) name) == 'p') {
        /* unnamed pipe */
        if (do_not_restore_flag) {
            retval = 0;
        } else {
            retval = cr_open_unnamed_pipe(req, file_info, cf_fifo);
        }
        if (retval < 0) {
            CR_ERR("cr_couple_fifo:  unable to open unnamed pipe: %s",
                   (char *) name);
            goto out;
        }
    } else {
        /* garbage */
        retval = -EINVAL;
        CR_ERR("bogus pipe filename in context file: %s", (char *) name);
        goto out;
    }


  out:
    return retval;
}

static int
cr_make_new_pipe(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	         struct cr_fifo *cf_fifo, char *name, int do_not_restore_flag)
{
    struct file *p_filp;
    int retval;

    CR_KTRACE_LOW_LVL("%s:  Phase 1: Making new pipe.", (char *) name);

    if (*((char *) name) == '/') {
        /* sanity check */
        if (!(S_IFIFO & cf_fifo->fifo_perms)) {
            CR_WARN("Pipe object not a FIFO!");
        }

        if (do_not_restore_flag) {
            retval = 0;
        } else {
            retval = cr_creat_named_fifo(file_info->fd, (char *) name,
                                         cf_fifo->fifo_flags,
                                         S_IFIFO | cf_fifo->fifo_perms,
					 file_info->unlinked);
        }

        if (retval < 0) {
            goto out;
        }
    } else if (*((char *) name) == 'p') {
        int pipe_fds[2];
        int tmp_fd;

        if (do_not_restore_flag) {
            retval = 0;
            goto out;
        }

        retval = do_pipe(pipe_fds);
        if (retval < 0) {
            goto out;
        }

        /* now have two pipe objects, close one of them. */
        if (cr_posix_flags_to_f_mode(cf_fifo->fifo_flags) == 1) {
            sys_close(pipe_fds[1]);
            tmp_fd = pipe_fds[0];
        } else if (cr_posix_flags_to_f_mode(cf_fifo->fifo_flags) == 2) {
            sys_close(pipe_fds[0]);
            tmp_fd = pipe_fds[1];
        } else {
            CR_ERR("Bad access mode for pipe.");
            goto out;
        }

        /* put the fd into the right place */
        if (tmp_fd != file_info->fd) {
            retval = sys_dup2(tmp_fd, file_info->fd);
            if (retval < 0) {
                CR_ERR("sys_dup2(%d,%d) failed.", tmp_fd, file_info->fd);
                sys_close(tmp_fd);
                /* we ignore the error from close here. */
                goto out;
            }

            /* we ignore the error from close here. */
            sys_close(tmp_fd);
        }

    } else {
        /* garbage */
        retval = -EINVAL;
        CR_ERR("Bogus pipe filename in context file: %s", (char *) name);
        goto out;
    }

    p_filp = fget(file_info->fd);
    if (!p_filp) {
        goto out;
    }

    /* now do the easy stuff */
    p_filp->f_pos = cf_fifo->fifo_pos;
    p_filp->f_flags = cf_fifo->fifo_flags;
    p_filp->f_version = cf_fifo->fifo_version;
    /* p_filp->f_fop was initialized correctly by do_pipe/mknod */

    /* fifo_id(inode), dentry, vfsmnt pointers into object table for subsequent open(s) */
    cr_insert_object(req->map, cf_fifo->fifo_id, p_filp->f_dentry->d_inode);
    cr_insert_object(req->map, cf_fifo->fifo_dentry, p_filp->f_dentry);
    cr_insert_object(req->map, cf_fifo->fifo_vfsmnt, p_filp->f_vfsmnt);

    fput(p_filp);

out:
    return retval;
}

static int 
cr_restore_open_fifo(cr_rstrt_req_t *req, struct cr_file_info *file_info,
       	int do_not_restore_flag)
{
    int retval = -ENOSYS;
    struct cr_fifo cf_fifo;
    struct file *cf_filp;
    struct file *p_filp;
    void *buf;
    int len;

    CR_KTRACE_FUNC_ENTRY();

    cf_filp = cr_src_gett(req, cr_fifo_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (void *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_putt;
    }

    /* read fifo struct */
    retval = cr_kread(cf_filp, &cf_fifo, sizeof(cf_fifo));
    if (retval != sizeof(cf_fifo)) {
        CR_ERR("cr_restore_open_fifo - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    CR_KTRACE_LOW_LVL("   Open fifo: id == %p.", cf_fifo.fifo_id);

    if (cf_fifo.cr_type != cr_fifo_obj) {
        CR_ERR
            ("cr_restore_open_fifo.  Garbage found in context file.  Wrong object type.");
        retval = -EINVAL;
        goto out_free;
    }

    /* read file name */
    retval = cr_fgets(buf, PAGE_SIZE, cf_filp, &len);
    if (retval < 0) {
        CR_ERR("pipe_name: Bad file read!");
        goto out_free;
    }

    /* now, actually rebuild the thing. */
    if (cf_fifo.fifo_len == ((unsigned int)-1)) {
        retval = cr_couple_fifo(req, file_info, &cf_fifo, buf,
                                do_not_restore_flag);
        if (retval < 0) {
            goto out_free;
        }

        /* 
         * the pipe buffer should have been restored when we first built the pipe
         */
    } else {
        retval = cr_make_new_pipe(req, file_info, &cf_fifo, buf,
                                  do_not_restore_flag);
        if (retval < 0) {
            goto out_free;
        }

	/* now read the pipe buffer data, and optionally restore it */
	p_filp = fget(file_info->fd);
	if (!p_filp) {
	    goto out_free;
	}
	retval = cr_restore_pipe_buf(cf_filp, p_filp->f_dentry->d_inode,
                                     cf_fifo.fifo_len, do_not_restore_flag);
	fput(p_filp);
	if (retval < 0) {
	    CR_ERR("could not restore pipe buffer");
	    goto out_free;
	}
    }

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_restore_open_socket(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    /* XXX:  Kludge to recover w/o restoring a socket */
    retval = 0;

    return retval;
}

static int 
cr_restore_open_chr(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_chrdev cf_chrdev;
    struct file *c_filp;

    CR_KTRACE_FUNC_ENTRY();

    c_filp = cr_src_gett(req, cr_chrdev_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_chrdev struct */
    retval = cr_kread(c_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
        CR_ERR("cr_restore_open_chr: Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* XXX:  Validate type */
    retval = -EINVAL;
    if (cf_chrdev.cr_type != cr_open_chr) {
        CR_ERR("cr_restore_open_chr: Garbage in context file.");
        goto out_putt;
    }

    CR_KTRACE_LOW_LVL("   Character device.  major=%d, minor=%d", cf_chrdev.cr_major, cf_chrdev.cr_minor);

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_blk(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int
cr_dup2(cr_rstrt_req_t *req, struct file *old_filp, int fd)
{
    struct file *new_filp;
    int retval = 0;

    /* translate the filp to its new value */
    if (!cr_find_object(req->map, old_filp, (void **) &new_filp)) {
        CR_ERR("cr_dup2: Could not locate restored filp %p.", old_filp);
	retval = -EINVAL;
        goto out;
    }
    get_file(new_filp);

    /* allocate the fd */
    retval = cr_fd_claim(fd);
    if (retval) {
        /* this should never happen.  */
        CR_ERR("cr_dup2: File descriptor %d in use.", fd);
	fput(new_filp);
        goto out;
    }

    /* install it */
    fd_install(fd, new_filp);

out:
    return retval;
}

static int 
cr_restore_open_dup(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_dup cf_dup;
    struct file *cf_filp;

    CR_KTRACE_FUNC_ENTRY();

    cf_filp = cr_src_gett(req, cr_fifo_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_dup struct */
    retval = cr_kread(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
        CR_ERR("cr_restore_open_dup: Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* Validate type (the only field we've got) */
    retval = -EINVAL;
    if (cf_dup.cr_type != cr_open_dup) {
        CR_ERR("cr_restore_open_dup: Garbage in context file.");
        goto out_putt;
    }

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = cr_dup2(req, file_info->orig_filp, file_info->fd);
    if (retval<0) {
        CR_ERR("cr_restore_open_dup: Could not cr_dup2(%p,%d).", file_info->orig_filp, file_info->fd);
        goto out_putt;
    }
    CR_KTRACE_LOW_LVL("cr_dup2(%d,%p) successful", file_info->fd, file_info->orig_filp);

    retval = 0;

out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_end_of_file_restore(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_restore_file_locks(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_finish_all_file_restores(cr_rstrt_req_t *req)
{
    int retval = -ENOSYS;
    int fd;

    CR_KTRACE_FUNC_ENTRY();

    cr_file_lock_read(&current->files->file_lock);
    CR_KTRACE_LOW_LVL("current->files=%p", current->files);
    for(fd=0; fd<current->files->max_fds; ++fd) {
      struct file *filp;
      filp = fcheck(fd);
      if (filp) {
        CR_KTRACE_LOW_LVL("fd=%d filp=%p filp->dentry=%p d_count=%d d_inode=%p", fd, filp, filp->f_dentry, atomic_read(&filp->f_dentry->d_count), filp->f_dentry->d_inode);
      }
    }
    cr_file_lock_unread(&current->files->file_lock);

    return retval;
}

/*
 * cr_restore_all_files
 *
 * Main entry point for file recovery
 *
 * In the future, may want to set this so that it manipulates file "objects"
 * with load / restore methods
 */
static int
cr_restore_all_files(cr_rstrt_req_t *req)
{
    int do_not_restore_flag;
    int i, retval;
    struct cr_file_info file_info;
    struct files_struct *files = current->files;

    /* load the files struct */
    CR_KTRACE_HIGH_LVL("recovering fs_struct...");
    retval = cr_restore_files_struct(req);
    if (retval) {
        CR_ERR("Unable to cr_restore_files_struct");
        goto out;
    }

    /* close-on-exec() of caller's fds before we start restoring */
    cr_file_lock_write(&files->file_lock);
    for (i = 0; (i < files->max_fds) && (i < files->max_fdset); ++i) {
	if (FD_ISSET(i, files->close_on_exec)) {
	    cr_file_lock_unwrite(&files->file_lock);
	    sys_close(i);
	    cr_file_lock_write(&files->file_lock);
        }
    }
    cr_file_lock_unwrite(&files->file_lock);

    file_info.cr_file_type = cr_bad_file_type;
    /* now do per file stuff */
    while (1) {
        retval = cr_load_file_info(req, &file_info);
	if (retval < 0) {
            CR_ERR("Error loading file_info.");
            goto out;
	}


	if (file_info.cr_file_type == cr_end_of_files) {
            /* finished */
            retval = cr_end_of_file_restore(req, &file_info);
            break;
	}

	cr_file_lock_read(&files->file_lock);
        do_not_restore_flag = FD_ISSET(file_info.fd, files->open_fds);
	cr_file_lock_unread(&files->file_lock);

        CR_KTRACE_LOW_LVL("   fd=%d flag=%d", file_info.fd, do_not_restore_flag);

        switch(file_info.cr_file_type) {
        case cr_open_file:
            retval = cr_restore_open_file(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_directory:
            retval = cr_restore_open_dir(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_link:
            retval = cr_restore_open_link(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_fifo:
            retval = cr_restore_open_fifo(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_socket:
            retval = cr_restore_open_socket(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_chr:
            retval = cr_restore_open_chr(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_blk:
            retval = cr_restore_open_blk(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_dup:
            retval = cr_restore_open_dup(req, &file_info, do_not_restore_flag);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            CR_ERR("Unknown file type fd=%d type=%d", file_info.fd, file_info.cr_file_type);
            retval = -EBADF;
            goto out;
            break;
        }

        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fd %d (type=%d,err=%d)",
                   __FUNCTION__, current->pid, file_info.fd, file_info.cr_file_type, retval);
            goto out;
        }

	/* Restore close-on-exec flag */
        cr_file_lock_write(&files->file_lock);
        if (file_info.cloexec) {
	    FD_SET(file_info.fd, files->close_on_exec);
        } else {
	    FD_CLR(file_info.fd, files->close_on_exec);
	}
	cr_file_lock_unwrite(&files->file_lock);

        /* now acquire locks (unimplemented) */
        cr_restore_file_locks(req, &file_info);

	cr_file_lock_read(&files->file_lock);
        cr_insert_object(req->map, file_info.orig_filp, fcheck(file_info.fd));
	cr_file_lock_unread(&files->file_lock);
    };

    /* now for the second pass.  The above restore routines should have taken
     * care of loading all the data.  the easy work, now we do everything 
     * that has to be deferred until we have all the data
     */
    cr_finish_all_file_restores(req);

out:
    return retval;
} 

/*
 * cr_chdir
 *
 * Our own version of the chdir() system call.  Differs in that it accepts
 * pathnames in kernel space, and calls path_init, path_walk, and path_release
 * directly.
 * 
 */
static int
cr_chdir(const char *path)
{
    struct nameidata nd;
    int retval;

    /* lookup the path */
#ifndef LOOKUP_POSITIVE
  /* XXX: is this correct? */
  #define LOOKUP_POSITIVE 0
#endif
    retval = cr_path_lookup(path,LOOKUP_POSITIVE|LOOKUP_FOLLOW|LOOKUP_DIRECTORY,&nd);
    if (retval)
        goto out;

    retval = cr_permission(nd.dentry->d_inode,MAY_EXEC,&nd);
    if (retval)
        goto dput_and_out;

    set_fs_pwd(current->fs, nd.mnt, nd.dentry);

dput_and_out:
    path_release(&nd);
out:
    return retval;
}

/*
 * cr_restore_fs
 *
 * This function does the work of restoring the data in our fs_struct, i.e.
 * the current working directory, and the umask.
 *
 * In the future, this may be modified to also restore a process after a
 * chroot(), if desirable.
 * 
 */
static int
cr_restore_fs_struct(cr_rstrt_req_t *req)
{
    int len;
    int retval;
    int umask;
    char *buf;
    struct file *filp;

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_nomem;
    }

    filp = cr_src_gett(req, cr_fs_obj);
    if (!filp) {
        retval = -EBADF;
        goto out_nofile;
    }

    /* restore the umask */
    retval = cr_kread(filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
        CR_ERR("umask: Bad file read! retval=%d", retval);
        retval = -EIO;
        goto out_readerr;
    }

    /* read in root */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("root: Bad file read!");
        goto out_readerr;
    }

    /* read in altroot */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("altroot: Bad file read!");
        goto out_readerr;
    } else if (retval > 0) {
        CR_ERR("Warning:  chroot recovery not implemented.");
    }

    /* read in the current working directory */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("pwd: Bad file read!");
        goto out_readerr;
    }

    /* should change root directory here. */
    /* cr_chroot() */

    /* now change the current working directory */
    retval = cr_chdir(buf);
    if (retval < 0) {
        CR_ERR("cr_chdir:  Unable to change directory to %s", buf);
	goto out_readerr;
    }    

out_readerr:
    cr_src_putt(filp);
out_nofile:
    free_page((unsigned long) buf);
out_nomem:
    return retval;
}

/*
 * This does half of a close.  It removes the fd from a files_struct, but doesn't
 * actually close the fd.
 */
int
cr_hide_filp(struct file *filp, struct files_struct *files)
{
    int found;
    int fd;

    found = 0;
    cr_file_lock_write(&files->file_lock);
    for (fd = 0; fd < files->max_fds; ++fd) {
	if (filp == files->fd[fd]) {
	    ++found;
	    files->fd[fd] = NULL;
	    FD_CLR(fd, files->close_on_exec);
	    break;
	}
    }
    cr_file_lock_unwrite(&files->file_lock);

    if (found) {
	put_unused_fd(fd);
    } else {
	CR_ERR("File descriptor closed unexpectedly!");
    }

    return found;
}

// cr_rstrt_clones
//
// Invoked by a freshly created child process.
//
// Returns clone flags if additional tasks must be spawned, 0 if not,
// or a negative value on failure.
// On success stores newly allocated stack pointer to the supplied address.
//
// XXX: how to keep ref counts?
int cr_rstrt_clones(struct file *filp, void **new_stack_p)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out;
    }

    down(&req->serial_mutex);
    if (req->clones_needed) {
	unsigned long new_stack;

        --req->clones_needed;

	/* allocate a new empty 1-page stack */
	cr_mmap_sem_down_write(&current->mm->mmap_sem);
	new_stack = do_mmap(0, 0, PAGE_SIZE,
			    PROT_READ | PROT_WRITE | PROT_EXEC,
			    MAP_PRIVATE | MAP_ANONYMOUS, 0);
	cr_mmap_sem_up_write(&current->mm->mmap_sem);
	if (!new_stack) {
	    retval = -ENOMEM;
	} else {
#if (CR_STACK_GROWTH < 0) /* Stack grows down */
	    new_stack += PAGE_SIZE;
#elif (CR_STACK_GROWTH > 0) /* Stack grows up */
	    /* Nothing to do */
#else
	    #error "Don't know which way the stack grows"
#endif
	    if (!(retval = put_user(new_stack, (unsigned long *)new_stack_p))) {
		/* success */
		retval = CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND;
#if HAVE_2_6_SIGNAL_STRUCT
		if (req->shared_signal) {
		    retval |= CLONE_THREAD | CLONE_DETACHED;
		}
#endif
	    }
	}
    } else {
        CR_KTRACE_HIGH_LVL("%d: Have enough processes", current->pid);
	retval = 0;
    }
    up(&req->serial_mutex);

out:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// cr_rstrt_child
//
// Invoked by a freshly created child process.
//
// The real "restart" part of reconstructing a process image happens here.
//
// Returns 1 on success, or a negative on failure
int cr_rstrt_child(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    cr_task_t *cr_task = NULL;
    int retval;
    int old_pid;
    long state = TASK_RUNNING;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out;
    }
    atomic_inc(&req->ref_count);
    clear_bit(0, &req->need_children);


    write_lock(&cr_task_lock);
    {
	cr_task = __cr_task_get(current, 1);
        if (!cr_task) {
		write_unlock(&cr_task_lock);
		retval = -ENOMEM;
                goto out;
        }
	cr_task->rstrt_req = req;
    }
    write_unlock(&cr_task_lock);

    /* Restore credentials (uids and gids) */
    down(&req->serial_mutex);
    if (!test_and_set_bit(0, &req->done_creds)) {
        CR_KTRACE_HIGH_LVL("%d: Restoring credentials", current->pid);
	retval = cr_restore_creds(req); 
    }
    up(&req->serial_mutex);

    /* Thaw thyself out (called by all threads) */
    {
        struct vmadump_threadundump_args thaw_args;
	long err;

	thaw_args.num_threads = req->total_threads;
	thaw_args.fd = -1;              /* Ignored, but we initialize */
	thaw_args.flags = 0;            /* no flags */
	thaw_args.old_pid = -1;         /* Output, but initialize anyway */

        err = vmadump_thaw_threads(req->file, get_pt_regs(current), &thaw_args);
	if (err < 0) {
	    CR_ERR("thaw_threads returned error, aborting. %ld", err);
	    req->die = 1;
            goto out_release;
	}
        old_pid = thaw_args.old_pid;

	retval = 0;
    }

    /* Restore linkage (PID, session ID, etc.) (called by all threads) */
    CR_KTRACE_HIGH_LVL("Formerly %s PID %d", current->comm, old_pid);
    retval = cr_restore_linkage(req, old_pid);
    if (retval < 0) {
        CR_ERR("%s [%d]:  Unable to restore linkage!  (err=%d)",
	       __FUNCTION__, current->pid, retval);
        // XXX:  Get the hell out of here...
        // XXX:  goto ok here?
	req->die = 1;
        goto out_release;
    }

    /* --- File recovery starts here --- */
    /* Exactly one thread should close the rstrt_req file descriptor(s)
     * but they should all wait until this has been done. */
    /* NOTE:  Past this point, you must jump to out_close to abort.  We are removing
     * our reference to the fd, which means do_exit() can't clean up after us. */ 

    down(&req->serial_mutex);
    if (!test_and_set_bit(0, &req->done_hide_cr_fds)) {
        /* Remove the rstrt_req descriptor from our open files. */
        cr_hide_filp(filp, current->files);
        cr_hide_filp(req->file, current->files);
    }

    /* restore fs_struct */
    /* XXX:  put the bit tests inside the restore routines when we do process
     * groups and sessions.  This way, they can just decide they want to skip it.
     */
    if (!test_and_set_bit(0, &req->done_fs)) {
        retval = cr_restore_fs_struct(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fs!  (err=%d)",
                   __FUNCTION__, current->pid, retval);

	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* restore the open files */
    if (!test_and_set_bit(0, &req->done_files)) {
        retval = cr_restore_all_files(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore files!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
            up(&req->serial_mutex);
	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* close the request descriptor */
out_close:
    if (!test_and_set_bit(0, &req->done_close_cr_fds)) {
	int rc;	/* don't overwrite error code w/ 0 */
	CR_KTRACE_LOW_LVL("%d: closing request descriptor", current->pid);
	rc = filp_close(filp, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing checkpoint control descriptor!  "
		   "(err=%d)", rc);
	    retval = rc;
	    req->die = 1;
	}

	CR_KTRACE_LOW_LVL("%d: closing context file descriptor", current->pid);
	rc = filp_close(req->file, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing context file descriptor!"
		   "(err=%d)", rc);
	    retval = rc;
	    req->die = 1;
	}
    }
    up(&req->serial_mutex);

    /* restore handlers (if any)  (all threads) */
    {
	/* XXX: move to cr_rstrt_task_complete()? */
	struct cr_context_tasklinkage *entry = cr_linkage_find_by_pid(req, current->pid);
	if ((entry != NULL) && (entry->cr_task.phase != CR_NO_PHASE)) {
    	    if (!cr_find_object(req->map, entry->cr_task.orig_filp, (void **) &cr_task->filp)) {
		CR_ERR("unable to locate checkpoint control descriptor");
	        req->die = 1;
                goto out_release;
	    }
	    cr_task->phase = entry->cr_task.phase;
	    cr_task->handler = entry->cr_task.handler;
	    atomic_inc(&cr_task->ref_count);
	}
    }
	    
out_release:
    /* On failure we need to ensure everyone fails, and inform the requester */
    if ((req->result >= 0) && (retval < 0)) {
	req->result = retval;
    }
    if (req->die) { /* set by us or another thread */
	send_sig_info(SIGKILL, NULL, current);
    }
    {
        /* XXX: move cr_rstrt_task_complete now that handler re-registration is done */
        if (req->linkage) {
	    struct cr_context_tasklinkage *entry = cr_linkage_find_by_pid(req, current->pid);
	    if (entry != NULL && entry->cr_task.stopped) {
	        state = TASK_STOPPED;
	    }
        }
	cr_task->rstrt_req = NULL;
	cr_task_put(cr_task);
        release_rstrt_req(req);
    }
out:
    if (retval == 0) {
        /* Return > zero for success restart */
	set_current_state(state);	/* XXX: move to cr_rstrt_task_complete? */
        retval = 1;
    }
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

/**
 * cr_rstrt_complete
 * @filp: The file on which the ioctl was received.
 *
 * DESCRIPTION:
 * Implements the HAND_RSTRT ioctl() which indicates to the kernel that
 * the calling task has completed running its RESTART handlers.  Thus
 * we can prevent checkpoint requests from overlapping these handlers.
 * (XXX: need to move stuff out of rstrt_child to get this right.)
 *
 * This is the function that the handlers will call into when
 * the RESTART portion of callbacks are finished.
 *
 */
int cr_rstrt_complete(struct file *filp)
{
#if 0
    cr_task_t *cr_task;

    CR_KTRACE_FUNC_ENTRY();

    // Lookup this task
    cr_task = cr_task_get(current);
    if (!cr_task) {
	// No matching task found.
	return -ESRCH;
    }

    // Find the matching request
    if(!cr_task->rstrt_req) {
	// No matching request found.
	cr_task_put(cr_task);
	return -ESRCH;
    }

    // Complete it
    cr_rstrt_task_complete(cr_task);

    // Release the task.
    cr_task_put(cr_task);
#endif

    return 0;
}

int cr_rstrt_req_release(struct file *file, cr_pdata_t *priv)
{
    CR_KTRACE_FUNC_ENTRY();
    if (priv == NULL) {
        CR_WARN("Warning:  Called with NULL priv!");
        return -EINVAL;
    }

    if (priv->rstrt_req) {
        release_rstrt_req(priv->rstrt_req);
    }

    return 0;
}
